@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# Copyright 2022 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -50,7 +50,7 @@ class IntegrationGenericTests(unittest.TestCase):
 
     TASK_NAME = 'noop'
 
-    def setUp(self):
+    def setUp(self) -> None:
         """Initialize test."""
         # If debusine-server or nginx was launched just before the
         # integration-tests.py is launched the debusine-server might not be
@@ -65,7 +65,7 @@ class IntegrationGenericTests(unittest.TestCase):
 
         self._worker = Worker()
 
-    def connect_worker(self):
+    def connect_worker(self) -> None:
         """
         Connect the worker to the debusine-server.
 
@@ -86,7 +86,7 @@ class IntegrationGenericTests(unittest.TestCase):
             DebusineServer.wait_for_worker_connected(Worker.read_token())
         )
 
-    def test_secret_key_file_not_readable_fails(self):
+    def test_secret_key_file_not_readable_fails(self) -> None:
         """Key file cannot be read: the server does not start."""
         key_file = "/var/lib/debusine/server/key"
         tmp_key_file = "/var/lib/debusine/server/key.tmp"
@@ -98,7 +98,7 @@ class IntegrationGenericTests(unittest.TestCase):
         # Cannot start because key file cannot be read
         self.assertEqual(DebusineServer.restart().returncode, 1)
 
-    def test_worker_is_not_connected(self):
+    def test_worker_is_not_connected(self) -> None:
         """The worker is not connected: the token is not enabled yet."""
         self._worker.set_up()
 
@@ -110,7 +110,7 @@ class IntegrationGenericTests(unittest.TestCase):
             )
         )
 
-    def test_worker_connects(self):
+    def test_worker_connects(self) -> None:
         """The admin enables the worker and it gets connected."""
         self.connect_worker()
 
@@ -120,7 +120,7 @@ class IntegrationGenericTests(unittest.TestCase):
             )
         )
 
-    def test_create_work_requests_invalid(self):
+    def test_create_work_requests_invalid(self) -> None:
         """Create an invalid work request."""
         # Client submits an invalid work-request (status will be 'error')
         # It is not valid because task_data is not valid
@@ -138,7 +138,7 @@ class IntegrationGenericTests(unittest.TestCase):
         self.assertIn("title", create_work_request_output["error"])
         self.assertIn("detail", create_work_request_output["error"])
 
-    def test_create_work_request_valid(self):
+    def test_create_work_request_valid(self) -> None:
         """Create a valid work request."""
         self.connect_worker()
 
@@ -154,7 +154,7 @@ class IntegrationGenericTests(unittest.TestCase):
             expected_result="success",
         )
 
-    def test_debusine_admin_print_permission_error_message(self):
+    def test_debusine_admin_print_permission_error_message(self) -> None:
         """debusine-admin is invoked with wrong user: print error message."""
         # This test is to catch invalid permission
         # (in the logs files for example). Since the test for the secret_key
@@ -176,8 +176,8 @@ class IntegrationGenericTests(unittest.TestCase):
         self.assertEqual(result.returncode, 3)
 
     def assert_create_work_request(
-        self, *, task_data, expected_status: str, expected_result: str
-    ):
+        self, *, task_data: str, expected_status: str, expected_result: str
+    ) -> None:
         """
         Submit via debusine client a work request to the server.
 
@@ -190,7 +190,7 @@ class IntegrationGenericTests(unittest.TestCase):
 
         # Work Request is registered
         self.assertEqual(create_work_request_output['result'], 'success')
-        self.assertTrue('registered' in create_work_request_output['message'])
+        self.assertIn('registered', create_work_request_output['message'])
 
         work_request_id = create_work_request_output['work_request_id']
 
@@ -218,7 +218,7 @@ class IntegrationGenericTests(unittest.TestCase):
         self.assertEqual(show_work_request_output["status"], expected_status)
         self.assertEqual(show_work_request_output["result"], expected_result)
 
-    def test_web_login(self):
+    def test_web_login(self) -> None:
         """
         Ensure the web interface is accessible after login.
 
@@ -226,24 +226,22 @@ class IntegrationGenericTests(unittest.TestCase):
         https://salsa.debian.org/freexian-team/debusine/-/issues/563
         """
         s = requests.session()
-        r = s.get(f"{Configuration.get_base_url()}/accounts/login")
+        r = s.get(f"{Configuration.get_base_url()}/-/login")
         html = lxml.html.fromstring(r.content)
         csrftoken = html.xpath(
             """//input[@name="csrfmiddlewaretoken"]/@value"""
         )[0]
-        with open(os.getenv("AUTOPKGTEST_TMP") + "/test-password.txt") as f:
+        with open(os.environ["AUTOPKGTEST_TMP"] + "/test-password.txt") as f:
             password = f.read()[:-1]
         r = s.post(
-            f"{Configuration.get_base_url()}/debusine/accounts/login/",
+            f"{Configuration.get_base_url()}/-/login/",
             {
                 "username": "test-user",
                 "password": password,
                 "csrfmiddlewaretoken": csrftoken,
             },
             # Referer required in https
-            headers={
-                "Referer": f"{Configuration.get_base_url()}/accounts/login"
-            },
+            headers={"Referer": f"{Configuration.get_base_url()}/-/login"},
         )
         html = lxml.html.fromstring(r.content)
         self.assertEqual(r.status_code, 200)
