@@ -1,4 +1,4 @@
-# Copyright 2022-2024 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -14,45 +14,46 @@ import http
 import logging
 import math
 import os
-import re
 import shutil
 import signal
 import subprocess
 import sys
 import tempfile
-from collections.abc import Callable
+import types
+from collections.abc import Generator
+from contextlib import contextmanager
 from datetime import datetime
-from functools import partial
 from itertools import islice
 from pathlib import Path
-from typing import Any
+from typing import Any, NoReturn
 
 import yaml
 from dateutil.parser import isoparse
-from debian.deb822 import Changes, Dsc
 from requests.exceptions import RequestException
 from yaml import YAMLError
 
-from debusine.artifacts import (
-    BinaryPackage,
-    BinaryPackages,
-    LocalArtifact,
-    SourcePackage,
-    Upload,
-)
+from debusine.artifacts import LocalArtifact, Upload
+from debusine.assets import AssetCategory, asset_data_model
 from debusine.client import exceptions
+from debusine.client.client_utils import (
+    get_debian_package,
+    prepare_changes_for_upload,
+    prepare_deb_for_upload,
+    prepare_dsc_for_upload,
+)
 from debusine.client.config import ConfigHandler
 from debusine.client.debusine import Debusine
 from debusine.client.exceptions import DebusineError
 from debusine.client.models import (
     CreateWorkflowRequest,
+    RelationType,
     WorkRequestExternalDebsignRequest,
     WorkRequestRequest,
     WorkRequestResponse,
     WorkflowTemplateRequest,
     model_to_json_serializable_dict,
+    model_to_json_serializable_list,
 )
-from debusine.client.utils import get_debian_package
 
 
 class Cli:
@@ -64,16 +65,16 @@ class Cli:
         main.execute()
     """
 
-    def __init__(self, argv):
+    def __init__(self, argv: list[str]) -> None:
         """Initialize object."""
         self._argv = argv
 
     @staticmethod
-    def _exit(signum, frame):
+    def _exit(signum: int, frame: types.FrameType | None) -> NoReturn:
         signum, frame  # fake usage for vulture
         raise SystemExit(0)
 
-    def _parse_args(self):
+    def _parse_args(self) -> None:
         """Parse argv and store results in self.args."""
         parser = argparse.ArgumentParser(
             prog='debusine',
@@ -334,6 +335,48 @@ class Cli:
             help="Show reverse relationships, rather than forward ones",
         )
 
+        create_asset = subparsers.add_parser(
+            "create-asset", help="Create an asset"
+        )
+        create_asset.add_argument(
+            "category",
+            type=AssetCategory,
+            help="Category of asset",
+        )
+        create_asset.add_argument(
+            "--workspace",
+            type=str,
+            required=True,
+            help="Workspace for this asset",
+        )
+        create_asset.add_argument(
+            "--data",
+            type=argparse.FileType("r"),
+            required=True,
+            help="File path (or - for stdin) to read the data "
+            "for the asset. YAML format",
+        )
+
+        list_assets = subparsers.add_parser("list-assets", help="List assets")
+        list_assets.add_argument(
+            "--id",
+            metavar="ID",
+            type=int,
+            help="Show an individual asset by ID.",
+        )
+        list_assets.add_argument(
+            "--workspace",
+            metavar="NAME",
+            type=str,
+            help="List all assets in workspace.",
+        )
+        list_assets.add_argument(
+            "--work-request",
+            metavar="ID",
+            type=int,
+            help="List assets created by work request.",
+        )
+
         on_work_request_completed = subparsers.add_parser(
             "on-work-request-completed",
             help="Execute a command when a work request is completed. "
@@ -361,9 +404,11 @@ class Cli:
             help="Path to the command to execute",
         )
 
+        subparsers.add_parser("setup", help="Guided setup of debusine-client")
+
         self.args = parser.parse_args(self._argv)
 
-    def _build_debusine_object(self):
+    def _build_debusine_object(self) -> Debusine:
         """Return the debusine object matching the command line parameters."""
         configuration = ConfigHandler(
             server_name=self.args.server,
@@ -391,7 +436,7 @@ class Cli:
             logger=logger,
         )
 
-    def _setup_http_logging(self):
+    def _setup_http_logging(self) -> None:
         if self.args.debug:
             """Do setup urllib3&requests traces."""
             # https://docs.python-requests.org/en/latest/api/#api-changes
@@ -415,17 +460,31 @@ class Cli:
             requests_log.setLevel(logging.DEBUG)
             requests_log.addHandler(sh)
 
-            def print_to_log(*args):
+            def print_to_log(*args: str) -> None:
                 requests_log.debug(" ".join(args))
 
-            http.client.print = print_to_log
+            setattr(http.client, "print", print_to_log)
 
-    def execute(self):
+    def execute(self) -> None:
         """Execute the command requested by the user."""
         signal.signal(signal.SIGINT, self._exit)
         signal.signal(signal.SIGTERM, self._exit)
 
         self._parse_args()
+
+        # Early dispatch for subcommands that do not need the Debusine object
+        # to operate
+        match getattr(self.args, 'sub-command'):
+            case "setup":
+                from debusine.client.setup import setup_server
+
+                setup_server(
+                    config_file_path=self.args.config_file,
+                    server=self.args.server,
+                    scope=self.args.scope,
+                )
+                return
+
         debusine = self._build_debusine_object()
         self._setup_http_logging()
 
@@ -489,6 +548,25 @@ class Cli:
                 self._show_artifact_relations(
                     debusine, self.args.id, self.args.reverse
                 )
+            case "create-asset":
+                data = self.args.data.read()
+                self.args.data.close()
+                if self.args.data == sys.stdin:
+                    print("---")  # avoid confusion between input and output
+
+                self._create_asset(
+                    debusine,
+                    self.args.category,
+                    self.args.workspace,
+                    data,
+                )
+            case "list-assets":
+                self._list_assets(
+                    debusine,
+                    asset=self.args.id,
+                    workspace=self.args.workspace,
+                    work_request=self.args.work_request,
+                )
             case "on-work-request-completed":
                 self._on_work_request_completed(
                     debusine,
@@ -505,7 +583,7 @@ class Cli:
                     f"Unexpected sub-command name: {unreachable}"
                 )
 
-    def execute_create_work_request(self, debusine):
+    def execute_create_work_request(self, debusine: Debusine) -> None:
         """Execute the work request action requested by the user."""
         data = self.args.data.read()
         self.args.data.close()
@@ -526,39 +604,38 @@ class Cli:
         )
 
     @staticmethod
-    def _print_yaml(dictionary):
-        """Print dictionary to stdout as yaml."""
-        output = yaml.safe_dump(dictionary, sort_keys=False, width=math.inf)
+    def _print_yaml(data: Any) -> None:
+        """Print data to stdout as yaml."""
+        output = yaml.safe_dump(data, sort_keys=False, width=math.inf)
         print(output, end="")
 
     @staticmethod
-    def _fail(error, *, summary=None):
+    def _fail(error: object, *, summary: str | None = None) -> NoReturn:
         print(error, file=sys.stderr)
         if summary is not None:
             print(summary, file=sys.stderr)
         raise SystemExit(3)
 
-    def _list_work_requests(self, debusine: Debusine, limit: int):
+    def _list_work_requests(self, debusine: Debusine, limit: int) -> None:
         """List up to limit most recent work requests."""
-        work_requests = self._api_call_or_fail(
-            lambda: list(islice(debusine.work_request_iter(), limit))
-        )
+        with self._api_call_or_fail():
+            work_requests = list(islice(debusine.work_request_iter(), limit))
         result = {
             "results": [work_request.dict() for work_request in work_requests]
         }
         self._print_yaml(result)
 
-    def _show_work_request(self, debusine, work_request_id):
+    def _show_work_request(
+        self, debusine: Debusine, work_request_id: int
+    ) -> None:
         """Print the task information for work_request_id."""
-        work_request = self._api_call_or_fail(
-            debusine.work_request_get, work_request_id
-        )
+        with self._api_call_or_fail():
+            work_request = debusine.work_request_get(work_request_id)
 
         artifacts_information = []
         for artifact_id in work_request.artifacts:
-            artifact_information = self._api_call_or_fail(
-                debusine.artifact_get, artifact_id
-            )
+            with self._api_call_or_fail():
+                artifact_information = debusine.artifact_get(artifact_id)
             artifacts_information.append(
                 model_to_json_serializable_dict(artifact_information)
             )
@@ -573,23 +650,25 @@ class Cli:
         debusine: Debusine,
         task_name: str,
         workspace: str,
-        task_data,
-        event_reactions,
-    ):
-        task_data = self._parse_yaml_data(task_data)
-        event_reactions = self._parse_yaml_data(event_reactions)
+        task_data: str,
+        event_reactions: str,
+    ) -> None:
+        parsed_task_data = self._parse_yaml_data(task_data)
+        parsed_event_reactions = self._parse_yaml_data(event_reactions)
 
         work_request = WorkRequestRequest(
             task_name=task_name,
             workspace=workspace,
-            task_data=task_data,
-            event_reactions=event_reactions,
+            task_data=parsed_task_data,
+            event_reactions=parsed_event_reactions,
         )
 
+        output: dict[str, Any]
         try:
-            work_request_created = self._api_call_or_fail(
-                debusine.work_request_create, work_request
-            )
+            with self._api_call_or_fail():
+                work_request_created = debusine.work_request_create(
+                    work_request
+                )
         except DebusineError as err:
             output = {"result": "failure", "error": err.asdict()}
         else:
@@ -613,11 +692,10 @@ class Cli:
         if priority_adjustment is None:
             self._fail("Error: no changes specified")
 
-        self._api_call_or_fail(
-            debusine.work_request_update,
-            work_request_id,
-            priority_adjustment=priority_adjustment,
-        )
+        with self._api_call_or_fail():
+            debusine.work_request_update(
+                work_request_id, priority_adjustment=priority_adjustment
+            )
 
     def _retry_work_request(
         self,
@@ -625,10 +703,8 @@ class Cli:
         work_request_id: int,
     ) -> None:
         """Retry an aborted/failed work request."""
-        self._api_call_or_fail(
-            debusine.work_request_retry,
-            work_request_id,
-        )
+        with self._api_call_or_fail():
+            debusine.work_request_retry(work_request_id)
 
     def _provide_signature_debsign(
         self,
@@ -639,14 +715,15 @@ class Cli:
         """Provide a work request with an external signature using `debsign`."""
         # Get a version of the work request with its dynamic task data
         # resolved.
-        work_request = self._api_call_or_fail(
-            debusine.work_request_external_debsign_get, work_request.id
-        )
+        with self._api_call_or_fail():
+            work_request = debusine.work_request_external_debsign_get(
+                work_request.id
+            )
         assert work_request.dynamic_task_data is not None
-        unsigned = self._api_call_or_fail(
-            debusine.artifact_get,
-            work_request.dynamic_task_data["unsigned_id"],
-        )
+        with self._api_call_or_fail():
+            unsigned = debusine.artifact_get(
+                work_request.dynamic_task_data["unsigned_id"]
+            )
 
         with tempfile.TemporaryDirectory(
             prefix="debusine-debsign-"
@@ -662,9 +739,8 @@ class Cli:
                     or name.endswith(".dsc")
                     or name.endswith(".buildinfo")
                 ):
-                    self._api_call_or_fail(
-                        debusine.download_artifact_file, unsigned, name, path
-                    )
+                    with self._api_call_or_fail():
+                        debusine.download_artifact_file(unsigned, name, path)
             # Upload artifacts are guaranteed to have exactly one .changes
             # file.
             [changes_path] = [
@@ -673,7 +749,7 @@ class Cli:
 
             # Call debsign.
             subprocess.run(
-                ["debsign", changes_path.name, *debsign_args],
+                ["debsign", "--re-sign", changes_path.name, *debsign_args],
                 cwd=tmp,
                 check=True,
             )
@@ -682,29 +758,27 @@ class Cli:
             signed_local = Upload.create(
                 changes_file=changes_path, allow_remote=True
             )
-            signed_remote = self._api_call_or_fail(
-                debusine.upload_artifact,
-                signed_local,
-                workspace=unsigned.workspace,
-            )
+            with self._api_call_or_fail():
+                signed_remote = debusine.upload_artifact(
+                    signed_local, workspace=unsigned.workspace
+                )
 
             # Complete the work request using the signed artifact.
-            self._api_call_or_fail(
-                debusine.work_request_external_debsign_complete,
-                work_request.id,
-                WorkRequestExternalDebsignRequest(
-                    signed_artifact=signed_remote.id
-                ),
-            )
+            with self._api_call_or_fail():
+                debusine.work_request_external_debsign_complete(
+                    work_request.id,
+                    WorkRequestExternalDebsignRequest(
+                        signed_artifact=signed_remote.id
+                    ),
+                )
 
     def _provide_signature(
         self, debusine: Debusine, work_request_id: int, extra_args: list[str]
     ) -> None:
         """Provide a work request with an external signature."""
         # Find out what kind of work request we're dealing with.
-        work_request = self._api_call_or_fail(
-            debusine.work_request_get, work_request_id
-        )
+        with self._api_call_or_fail():
+            work_request = debusine.work_request_get(work_request_id)
         match (work_request.task_type, work_request.task_name):
             case "Wait", "externaldebsign":
                 self._provide_signature_debsign(
@@ -737,10 +811,12 @@ class Cli:
             priority=priority,
         )
 
+        output: dict[str, Any]
         try:
-            workflow_template_created = self._api_call_or_fail(
-                debusine.workflow_template_create, workflow_template
-            )
+            with self._api_call_or_fail():
+                workflow_template_created = debusine.workflow_template_create(
+                    workflow_template
+                )
         except DebusineError as err:
             output = {"result": "failure", "error": err.asdict()}
         else:
@@ -786,10 +862,10 @@ class Cli:
             task_data=parsed_task_data,
         )
 
+        output: dict[str, Any]
         try:
-            workflow_created = self._api_call_or_fail(
-                debusine.workflow_create, workflow
-            )
+            with self._api_call_or_fail():
+                workflow_created = debusine.workflow_create(workflow)
         except DebusineError as err:
             output = {"result": "failure", "error": err.asdict()}
         else:
@@ -817,9 +893,10 @@ class Cli:
             task_data=data,
         )
 
-    def _api_call_or_fail(self, method, *args, **kwargs):
+    @contextmanager
+    def _api_call_or_fail(self) -> Generator[None, None, None]:
         """
-        Call method with args and kwargs.
+        Context manager to handle failures from calling a method.
 
         :raises: exceptions.NotFoundError: server returned 404.
         :raises: UnexpectedResponseError: e.g. invalid JSON.
@@ -828,7 +905,7 @@ class Cli:
           reports an error.
         """
         try:
-            result = method(*args, **kwargs)
+            yield
         except exceptions.NotFoundError as exc:
             self._fail(exc)
         except exceptions.UnexpectedResponseError as exc:
@@ -838,8 +915,6 @@ class Cli:
         except exceptions.ClientConnectionError as client_error:
             self._fail(f'Error connecting to debusine: {client_error}')
 
-        return result
-
     @classmethod
     def _parse_yaml_data(cls, data_yaml: str | None) -> dict[str, Any]:
         if data_yaml is not None:
@@ -861,18 +936,19 @@ class Cli:
         else:
             data = {}
 
+        assert isinstance(data, dict)
         return data
 
     def _create_artifact(
         self,
-        debusine,
+        debusine: Debusine,
         artifact_type: str,
         workspace: str,
         upload_paths: list[str],
         upload_base_directory: str | None,
         expire_at: datetime | None,
-        data_yaml,
-    ):
+        data_yaml: str | None,
+    ) -> None:
         data = self._parse_yaml_data(data_yaml)
 
         artifact_cls = LocalArtifact.class_from_category(artifact_type)
@@ -891,13 +967,12 @@ class Cli:
         except ValueError as exc:
             self._fail(f"Cannot create artifact: {exc}")
 
+        output: dict[str, Any]
         try:
-            artifact_created = self._api_call_or_fail(
-                debusine.artifact_create,
-                local_artifact,
-                workspace=workspace,
-                expire_at=expire_at,
-            )
+            with self._api_call_or_fail():
+                artifact_created = debusine.artifact_create(
+                    local_artifact, workspace=workspace, expire_at=expire_at
+                )
         except DebusineError as err:
             output = {"result": "failure", "error": err.asdict()}
             self._print_yaml(output)
@@ -919,97 +994,8 @@ class Cli:
                 ]
 
             self._print_yaml(output)
-            self._api_call_or_fail(
-                debusine.upload_files,
-                artifact_created.id,
-                files_to_upload,
-            )
-
-    def _prepare_changes_for_upload(
-        self, changes: Path
-    ) -> list[LocalArtifact[Any]]:
-        with changes.open("rb") as f:
-            parsed = Changes(f)
-        architectures = parsed["Architecture"].split()
-        artifact_factories: list[Callable[[], LocalArtifact[Any]]] = []
-        for architecture in architectures:
-            if architecture == "source":
-                dscs = [
-                    file["name"]
-                    for file in parsed["Files"]
-                    if file["name"].endswith(".dsc")
-                ]
-                if len(dscs) != 1:
-                    self._fail(
-                        f"Expecting exactly 1 .dsc in source package, "
-                        f"found {len(dscs)}"
-                    )
-                artifact_factories.append(
-                    partial(
-                        self._prepare_dsc_for_upload, changes.parent / dscs[0]
-                    )
-                )
-            else:
-                debs = [
-                    changes.parent / file["name"]
-                    for file in parsed["Files"]
-                    if file["name"].endswith(
-                        (
-                            f"_{architecture}.deb",
-                            f"_{architecture}.udeb",
-                        )
-                    )
-                ]
-                if len(debs) < 1:
-                    self._fail(
-                        "Expecting at least one .deb per arch in "
-                        "binary packages"
-                    )
-
-                srcpkg_name = parsed["Source"]
-                srcpkg_version = parsed["Version"]
-                m = re.match(
-                    r"^(?P<name>\S+) \((?P<version>\S+)\)$", srcpkg_name
-                )
-                if m:
-                    srcpkg_name = m.group("name")
-                    srcpkg_version = m.group("version")
-
-                # Create artifacts for each individual .deb.
-                for deb in debs:
-                    artifact_factories.append(
-                        partial(BinaryPackage.create, file=deb)
-                    )
-                # For convenience, also create a grouped artifact for all
-                # the .debs.
-                artifact_factories.append(
-                    partial(
-                        BinaryPackages.create,
-                        srcpkg_name=srcpkg_name,
-                        srcpkg_version=srcpkg_version,
-                        version=parsed["Version"],
-                        architecture=architecture,
-                        files=debs,
-                    )
-                )
-        artifacts: list[LocalArtifact[Any]] = [
-            Upload.create(changes_file=changes)
-        ] + [factory() for factory in artifact_factories]
-        return artifacts
-
-    def _prepare_dsc_for_upload(self, dsc: Path) -> SourcePackage:
-        files = [dsc]
-        directory = dsc.parent
-        with dsc.open() as f:
-            parsed = Dsc(f)
-            name = parsed["source"]
-            version = parsed["version"]
-            for file in parsed["files"]:
-                files.append(directory / file["name"])
-        return SourcePackage.create(name=name, version=version, files=files)
-
-    def _prepare_deb_for_upload(self, deb: Path) -> BinaryPackage:
-        return BinaryPackage.create(file=deb)
+            with self._api_call_or_fail():
+                debusine.upload_files(artifact_created.id, files_to_upload)
 
     def _import_debian_artifact(
         self,
@@ -1017,18 +1003,18 @@ class Cli:
         workspace: str,
         upload: str,
         expire_at: datetime | None,
-    ):
+    ) -> None:
         result: dict[str, Any] = {}
         uploaded = []
         try:
             with get_debian_package(upload) as path:
                 artifacts: list[LocalArtifact[Any]]
                 if path.suffix == ".changes":
-                    artifacts = self._prepare_changes_for_upload(path)
+                    artifacts = prepare_changes_for_upload(path)
                 elif path.suffix == ".dsc":
-                    artifacts = [self._prepare_dsc_for_upload(path)]
+                    artifacts = [prepare_dsc_for_upload(path)]
                 elif path.suffix in (".deb", ".ddeb", ".udeb"):
-                    artifacts = [self._prepare_deb_for_upload(path)]
+                    artifacts = [prepare_deb_for_upload(path)]
                 else:
                     raise ValueError(
                         "Only source packages (.dsc), binary packages (.deb), "
@@ -1036,12 +1022,10 @@ class Cli:
                         f"imported with this command. {path} is not supported."
                     )
                 for artifact in artifacts:
-                    remote_artifact = self._api_call_or_fail(
-                        debusine.upload_artifact,
-                        artifact,
-                        workspace=workspace,
-                        expire_at=expire_at,
-                    )
+                    with self._api_call_or_fail():
+                        remote_artifact = debusine.upload_artifact(
+                            artifact, workspace=workspace, expire_at=expire_at
+                        )
                     uploaded.append(remote_artifact)
             primary_artifact = uploaded[0]
             result = {
@@ -1054,21 +1038,21 @@ class Cli:
                 "artifact_id": primary_artifact.id,
             }
             for related_artifact in uploaded[1:]:
-                self._api_call_or_fail(
-                    debusine.relation_create,
-                    primary_artifact.id,
-                    related_artifact.id,
-                    "extends",
-                )
+                with self._api_call_or_fail():
+                    debusine.relation_create(
+                        primary_artifact.id,
+                        related_artifact.id,
+                        RelationType.EXTENDS,
+                    )
                 result.setdefault("extends", []).append(
                     {"artifact_id": related_artifact.id}
                 )
-                self._api_call_or_fail(
-                    debusine.relation_create,
-                    primary_artifact.id,
-                    related_artifact.id,
-                    "relates-to",
-                )
+                with self._api_call_or_fail():
+                    debusine.relation_create(
+                        primary_artifact.id,
+                        related_artifact.id,
+                        RelationType.RELATES_TO,
+                    )
                 result.setdefault("relates_to", []).append(
                     {"artifact_id": related_artifact.id}
                 )
@@ -1091,36 +1075,88 @@ class Cli:
         *,
         target_directory: Path,
         tarball: bool,
-    ):
+    ) -> None:
         if not os.access(target_directory, os.W_OK):
             self._fail(f"Error: Cannot write to {target_directory}")
 
-        self._api_call_or_fail(
-            debusine.download_artifact,
-            artifact_id,
-            destination=target_directory,
-            tarball=tarball,
-        )
+        with self._api_call_or_fail():
+            debusine.download_artifact(
+                artifact_id, destination=target_directory, tarball=tarball
+            )
 
-    def _show_artifact(self, debusine: Debusine, artifact_id: int):
-        artifact = self._api_call_or_fail(debusine.artifact_get, artifact_id)
+    def _show_artifact(self, debusine: Debusine, artifact_id: int) -> None:
+        with self._api_call_or_fail():
+            artifact = debusine.artifact_get(artifact_id)
         self._print_yaml(model_to_json_serializable_dict(artifact))
 
     def _show_artifact_relations(
         self, debusine: Debusine, artifact_id: int, reverse: bool
     ) -> None:
-        if reverse:
-            models = self._api_call_or_fail(
-                debusine.relation_list, target_id=artifact_id
-            )
-        else:
-            models = self._api_call_or_fail(
-                debusine.relation_list, artifact_id=artifact_id
-            )
+        with self._api_call_or_fail():
+            if reverse:
+                models = debusine.relation_list(target_id=artifact_id)
+            else:
+                models = debusine.relation_list(artifact_id=artifact_id)
 
         relations = [model_to_json_serializable_dict(model) for model in models]
         self._print_yaml(relations)
 
+    def _create_asset(
+        self,
+        debusine: Debusine,
+        category: AssetCategory,
+        workspace: str,
+        data_yaml: str | None,
+    ) -> None:
+        data = self._parse_yaml_data(data_yaml)
+
+        asset_data = asset_data_model(category, data)
+
+        try:
+            with self._api_call_or_fail():
+                asset = debusine.asset_create(
+                    category=category, data=asset_data, workspace=workspace
+                )
+        except DebusineError as err:
+            output = {"result": "failure", "error": err.asdict()}
+            self._print_yaml(output)
+            raise SystemExit(3)
+        else:
+            self._print_yaml(
+                {
+                    "result": "success",
+                    "message": (
+                        f"New asset created in {debusine.base_api_url} "
+                        f"in workspace {asset.workspace} with id {asset.id}."
+                    ),
+                }
+            )
+
+    def _list_assets(
+        self,
+        debusine: Debusine,
+        asset: int | None,
+        workspace: str | None,
+        work_request: int | None,
+    ) -> None:
+        if asset is None and workspace is None and work_request is None:
+            self._fail(
+                "At least one of --id, --workspace, and --work-request must "
+                "be specified."
+            )
+        try:
+            with self._api_call_or_fail():
+                assets = debusine.asset_list(
+                    asset_id=asset,
+                    work_request=work_request,
+                    workspace=workspace,
+                )
+        except DebusineError as err:
+            output = {"result": "failure", "error": err.asdict()}
+            self._print_yaml(output)
+            raise SystemExit(3)
+        self._print_yaml(model_to_json_serializable_list(assets))
+
     def _on_work_request_completed(
         self,
         debusine: Debusine,
@@ -1128,7 +1164,7 @@ class Cli:
         workspaces: list[str],
         last_completed_at: Path | None,
         command: str,
-    ):
+    ) -> None:
         if shutil.which(command) is None:
             self._fail(
                 f'Error: "{command}" does not exist or is not executable'
