@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# Copyright 2022-2024 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -37,11 +37,16 @@ from requests_toolbelt.downloadutils.stream import stream_response_to_file
 
 from debusine.artifacts import LocalArtifact
 from debusine.artifacts.models import CollectionCategory
+from debusine.assets import AssetCategory, BaseAssetDataModel
 from debusine.client.debusine_http_client import DebusineHttpClient
 from debusine.client.exceptions import TokenDisabledError
 from debusine.client.file_uploader import FileUploader
 from debusine.client.models import (
     ArtifactResponse,
+    AssetCreateRequest,
+    AssetPermissionCheckResponse,
+    AssetResponse,
+    AssetsResponse,
     CreateWorkflowRequest,
     LookupChildType,
     LookupMultipleRequest,
@@ -52,8 +57,8 @@ from debusine.client.models import (
     LookupSingleResponseCollection,
     OnWorkRequestCompleted,
     RelationCreateRequest,
-    RelationCreateRequestType,
     RelationResponse,
+    RelationType,
     RelationsResponse,
     RemoteArtifact,
     WorkRequestExternalDebsignRequest,
@@ -61,10 +66,11 @@ from debusine.client.models import (
     WorkRequestResponse,
     WorkflowTemplateRequest,
     WorkflowTemplateResponse,
+    model_to_json_serializable_dict,
 )
 
 if TYPE_CHECKING:
-    from debusine.tasks.models import LookupMultiple, LookupSingle
+    from debusine.tasks.models import LookupMultiple, LookupSingle, OutputData
 
 
 class _MessageProcessor(Protocol):
@@ -159,11 +165,18 @@ class Debusine:
         """Return path to update the completed result for work_request_id."""
         return f"/work-request/{work_request_id}/completed/"
 
-    def work_request_completed_update(self, work_request_id: int, result: str):
+    def work_request_completed_update(
+        self, work_request_id: int, result: str, output_data: "OutputData"
+    ) -> None:
         """Update work_request_id as completed with result."""
         return self._debusine_http_client.put(
             path=self.work_request_completed_path(work_request_id),
-            data={"result": result},
+            data={
+                "result": result,
+                "output_data": model_to_json_serializable_dict(
+                    output_data, exclude_unset=True
+                ),
+            },
             expected_class=None,
         )
 
@@ -317,6 +330,11 @@ class Debusine:
         :param base_directory: base directory for relative path's files
          to upload.
         """
+        # The file uploader won't work as an anonymous user.  If this
+        # assertion is optimized out then the file upload will fail
+        # harmlessly; this just helps with type-checking.
+        assert self.token is not None
+
         file_uploader = FileUploader(self.token)
 
         for artifact_path, local_path in upload_files.items():
@@ -368,7 +386,7 @@ class Debusine:
         self,
         artifact_id: int,
         target_id: int,
-        relation_type: RelationCreateRequestType,
+        relation_type: RelationType,
     ) -> RelationResponse:
         """
         Create a new relation between artifacts.
@@ -468,6 +486,9 @@ class Debusine:
         if not tarball and not self._should_download_as_tarball(artifact):
             self._logger.info("Downloading artifact files into %s", destination)
             for name, file in artifact.files.items():
+                self._logger.debug(
+                    "Downloading artifact file %s from %s", name, file.url
+                )
                 path = destination / name
                 if path.parent != destination:
                     if not path.is_relative_to(destination):
@@ -516,8 +537,7 @@ class Debusine:
             )
 
             assert proc.stdin is not None
-            # https://github.com/python/mypy/issues/14943
-            shutil.copyfileobj(raw_response, proc.stdin)  # type: ignore
+            shutil.copyfileobj(raw_response, proc.stdin)
 
             proc.stdin.close()
 
@@ -570,6 +590,76 @@ class Debusine:
 
         return artifact
 
+    def asset_create(
+        self,
+        category: AssetCategory,
+        data: BaseAssetDataModel,
+        *,
+        workspace: str,
+        work_request: int | None = None,
+    ) -> AssetResponse:
+        """Create asset in the debusine server."""
+        return self._debusine_http_client.post(
+            "/asset/",
+            AssetResponse,
+            data=json.loads(
+                AssetCreateRequest(
+                    category=category,
+                    workspace=workspace,
+                    work_request=work_request,
+                    data=data,
+                ).json()
+            ),
+            expected_statuses=[requests.codes.created],
+        )
+
+    def asset_list(
+        self,
+        *,
+        asset_id: int | None = None,
+        work_request: int | None = None,
+        workspace: str | None = None,
+    ) -> AssetsResponse:
+        """Create asset in the debusine server."""
+        params: dict[str, str | int] = {}
+        if asset_id:
+            params["asset"] = asset_id
+        if work_request:
+            params["work_request"] = work_request
+        if workspace:
+            params["workspace"] = workspace
+        if not params:
+            raise ValueError(
+                "At least one of asset, work_request, and workspace must be "
+                "specified"
+            )
+        query = urlencode(params)
+        return self._debusine_http_client.get(
+            f"/asset/?{query}",
+            AssetsResponse,
+        )
+
+    def asset_permission_check(
+        self,
+        *,
+        asset_category: str,
+        asset_slug: str,
+        permission_name: str,
+        artifact_id: int,
+        work_request_id: int,
+        workspace: str,
+    ) -> AssetPermissionCheckResponse:
+        """Check work_request's permission to use an asset on an artifact."""
+        return self._debusine_http_client.post(
+            f"/asset/{asset_category}/{asset_slug}/{permission_name}/",
+            AssetPermissionCheckResponse,
+            data={
+                "artifact_id": artifact_id,
+                "work_request_id": work_request_id,
+                "workspace": workspace,
+            },
+        )
+
     @overload
     def lookup_single(
         self,
@@ -660,7 +750,9 @@ class Debusine:
             "/lookup/multiple/", LookupMultipleResponse, data=request.dict()
         )
 
-    def _log_tenacity_exception(self, retry_state: tenacity.RetryCallState):
+    def _log_tenacity_exception(
+        self, retry_state: tenacity.RetryCallState
+    ) -> None:
         # outcome is always non-None in an "after" hook.
         assert retry_state.outcome is not None
 
@@ -673,8 +765,8 @@ class Debusine:
         command: str | os.PathLike[str],
         working_directory: Path,
         last_completed_at: Path | None,
-        msg_content,
-    ):
+        msg_content: dict[str, Any],
+    ) -> None:
         on_work_request_completed = OnWorkRequestCompleted.parse_obj(
             msg_content
         )
@@ -701,7 +793,7 @@ class Debusine:
         command: str | os.PathLike[str],
         working_directory: Path,
         last_completed_at: Path | None,
-    ):
+    ) -> None:
         """
         Connect using websockets to URL and wait for msg. Execute command.
 
@@ -717,7 +809,7 @@ class Debusine:
             ),
             after=self._log_tenacity_exception,
         )
-        async def do_wait_for_messages(session: aiohttp.ClientSession):
+        async def do_wait_for_messages(session: aiohttp.ClientSession) -> None:
             headers = {}
             if self.token is not None:
                 headers["Token"] = self.token
@@ -778,7 +870,9 @@ class Debusine:
             return None
 
         last_completed_at_json = json.loads(last_completed_at.read_bytes())
-        return last_completed_at_json.get("last_completed_at")
+        completed_at = last_completed_at_json.get("last_completed_at")
+        assert completed_at is None or isinstance(completed_at, str)
+        return completed_at
 
     def on_work_request_completed(
         self,
@@ -787,7 +881,7 @@ class Debusine:
         last_completed_at: Path | None = None,
         command: str,
         working_directory: Path,
-    ):
+    ) -> None:
         """Execute command when a work request is completed."""
         debusine_ws_url = self.base_api_url.replace("http", "ws", 1)
         url_work_request_completed = (
@@ -816,7 +910,7 @@ class Debusine:
             )
         )
 
-    async def _tenacity_sleep(self, delay: float):
+    async def _tenacity_sleep(self, delay: float) -> None:
         # Used for the unit tests
         await asyncio.sleep(delay)
 
