@@ -1,4 +1,4 @@
-# Copyright 2024 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -20,6 +20,7 @@ from debusine.artifacts.models import (
     ArtifactCategory,
     BareDataCategory,
     CollectionCategory,
+    DebusinePromise,
 )
 from debusine.client.models import LookupChildType
 from debusine.db.context import context
@@ -58,10 +59,9 @@ class LookupMixin(TestCase):
         """Set up common data for tests."""
         super().setUpTestData()
         cls.user = cls.playground.get_default_user()
-        cls.owners = Group.objects.create(
-            scope=cls.playground.default_scope, name="Owners"
+        cls.owners = cls.playground.create_group(
+            name="Owners", users=[cls.user]
         )
-        cls.owners.users.add(cls.user)
 
     def create_artifact_item(
         self,
@@ -78,7 +78,10 @@ class LookupMixin(TestCase):
             name=name,
             child_type=CollectionItem.Types.ARTIFACT,
             category=category,
-            artifact=artifact or self.create_artifact(category=category)[0],
+            artifact=(
+                artifact
+                or self.playground.create_artifact(category=category)[0]
+            ),
             data=data or {},
             created_by_user=self.user,
         )
@@ -107,7 +110,7 @@ class LookupMixin(TestCase):
 
     def create_source_package(self, name: str, version: str) -> Artifact:
         """Create a minimal `debian:source-package` artifact."""
-        artifact, _ = self.create_artifact(
+        artifact, _ = self.playground.create_artifact(
             category=ArtifactCategory.SOURCE_PACKAGE,
             data={
                 "name": name,
@@ -122,7 +125,7 @@ class LookupMixin(TestCase):
         self, name: str, version: str, architecture: str
     ) -> Artifact:
         """Create a minimal `debian:binary-package` artifact."""
-        artifact, _ = self.create_artifact(
+        artifact, _ = self.playground.create_artifact(
             category=ArtifactCategory.BINARY_PACKAGE,
             data={
                 "srcpkg_name": name,
@@ -213,18 +216,20 @@ class LookupSingleTests(LookupMixin, TestCase):
         with self.assertRaisesRegex(LookupError, "Empty lookup"):
             lookup_single("", default_workspace(), user=AnonymousUser())
 
+    @context.disable_permission_checks()
     def test_specific_artifact(self) -> None:
         """`nnn@artifacts` finds a specific artifact."""
-        artifacts = [self.create_artifact()[0] for _ in range(3)]
+        artifacts = [self.playground.create_artifact()[0] for _ in range(3)]
 
         for i, artifact in enumerate(artifacts):
             lookup = f"{artifact.id}@artifacts"
             with self.subTest(lookup=lookup):
                 self.assert_lookup_equal(lookup, artifact)
 
+    @context.disable_permission_checks()
     def test_specific_artifact_integer(self) -> None:
         """Integer lookups find a specific artifact."""
-        artifacts = [self.create_artifact()[0] for _ in range(3)]
+        artifacts = [self.playground.create_artifact()[0] for _ in range(3)]
 
         for i, artifact in enumerate(artifacts):
             with self.subTest(artifact=artifact):
@@ -248,15 +253,15 @@ class LookupSingleTests(LookupMixin, TestCase):
             private_workspaces = []
             for i in range(2):
                 workspace = self.playground.create_workspace(name=f"private{i}")
-                workspace.assign_role(Workspace.Roles.OWNER, self.owners)
+                self.owners.assign_role(workspace, "owner")
                 private_workspaces.append(workspace)
             public_workspace = self.playground.create_workspace(
                 name="public", public=True
             )
-        artifacts = [
-            self.create_artifact(workspace=workspace)[0]
-            for workspace in private_workspaces + [public_workspace]
-        ]
+            artifacts = [
+                self.playground.create_artifact(workspace=workspace)[0]
+                for workspace in private_workspaces + [public_workspace]
+            ]
 
         for artifact, workspace, visible in (
             (artifacts[0], private_workspaces[0], True),
@@ -322,7 +327,7 @@ class LookupSingleTests(LookupMixin, TestCase):
             private_workspaces = []
             for i in range(2):
                 workspace = self.playground.create_workspace(name=f"private{i}")
-                workspace.assign_role(Workspace.Roles.OWNER, self.owners)
+                self.owners.assign_role(workspace, "owner")
                 private_workspaces.append(workspace)
             public_workspace = self.playground.create_workspace(
                 name="public", public=True
@@ -371,7 +376,7 @@ class LookupSingleTests(LookupMixin, TestCase):
             wprivate = self.playground.create_workspace(
                 name="private", public=False
             )
-            wprivate.assign_role(Workspace.Roles.OWNER, self.owners)
+            self.owners.assign_role(wprivate, "owner")
             cprivate = self.playground.create_collection(
                 "test", CollectionCategory.TEST, workspace=wprivate
             )
@@ -464,7 +469,7 @@ class LookupSingleTests(LookupMixin, TestCase):
             private_workspaces: list[Workspace] = []
             for i in range(2):
                 workspace = self.playground.create_workspace(name=f"private{i}")
-                workspace.assign_role(Workspace.Roles.OWNER, self.owners)
+                self.owners.assign_role(workspace, "owner")
                 private_workspaces.append(workspace)
             public_workspace = self.playground.create_workspace(
                 name="public", public=True
@@ -522,6 +527,7 @@ class LookupSingleTests(LookupMixin, TestCase):
         ):
             lookup_single("debian", default_workspace(), user=self.user)
 
+    @context.disable_permission_checks()
     def test_collection_member(self) -> None:
         """Looking up collection members works."""
         self.playground.create_collection("bullseye", CollectionCategory.SUITE)
@@ -559,6 +565,7 @@ class LookupSingleTests(LookupMixin, TestCase):
                 user=self.user,
             )
 
+    @context.disable_permission_checks()
     def test_no_lookup_through_artifact(self) -> None:
         """Lookup segments cannot continue after an artifact."""
         suite = self.playground.create_collection(
@@ -583,6 +590,7 @@ class LookupSingleTests(LookupMixin, TestCase):
                 user=self.user,
             )
 
+    @context.disable_permission_checks()
     def test_expect_type(self) -> None:
         """`expect_type` requires the result to be of that type."""
         collection = self.playground.create_collection(
@@ -673,9 +681,10 @@ class LookupMultipleTests(LookupMixin, TestCase):
                 **kwargs,
             )
 
+    @context.disable_permission_checks()
     def test_not_collection(self) -> None:
         """The `collection` key does not identify a collection."""
-        artifact = self.create_artifact()[0]
+        artifact = self.playground.create_artifact()[0]
         lookup = {"collection": f"{artifact.id}@artifacts"}
 
         with self.assertRaisesRegex(
@@ -689,6 +698,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
                 user=AnonymousUser(),
             )
 
+    @context.disable_permission_checks()
     def test_child_type_bare(self) -> None:
         """Restrict to bare items."""
         collection = self.playground.create_collection(
@@ -711,6 +721,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
             (bare1, bare2),
         )
 
+    @context.disable_permission_checks()
     def test_child_type_artifact(self) -> None:
         """Restrict to artifacts."""
         bookworm = self.playground.create_collection(
@@ -725,7 +736,14 @@ class LookupMultipleTests(LookupMixin, TestCase):
             bookworm, "bare1", category=BareDataCategory.TEST
         )
         bare2 = self.playground.create_bare_data_item(
-            bookworm, "bare2", category=BareDataCategory.PROMISE
+            bookworm,
+            "bare2",
+            category=BareDataCategory.PROMISE,
+            data=DebusinePromise(
+                promise_work_request_id=2,
+                promise_workflow_id=1,
+                promise_category=ArtifactCategory.TEST,
+            ),
         )
         self.create_collection_item(bookworm, "collection")
         items = []
@@ -758,6 +776,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
             {"collection": "bookworm@debian:suite"}, items[:2]
         )
 
+    @context.disable_permission_checks()
     def test_child_type_collection(self) -> None:
         """Restrict to collections."""
         collection = self.playground.create_collection(
@@ -780,6 +799,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
             (collection1, collection2),
         )
 
+    @context.disable_permission_checks()
     def test_child_type_any(self) -> None:
         """Don't restrict by child type."""
         collection = self.playground.create_collection(
@@ -847,6 +867,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
                             expect_type=expect_type,
                         )
 
+    @context.disable_permission_checks()
     def test_category(self) -> None:
         """Restrict by category."""
         bookworm = self.playground.create_collection(
@@ -892,6 +913,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
             items[2:],
         )
 
+    @context.disable_permission_checks()
     def test_name_matcher(self) -> None:
         """Restrict by name."""
         collection = self.playground.create_collection(
@@ -931,6 +953,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
             items[:2],
         )
 
+    @context.disable_permission_checks()
     def test_data_matchers(self) -> None:
         """Restrict by data."""
         collection = self.playground.create_collection(
@@ -989,6 +1012,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
             (items[2],),
         )
 
+    @context.disable_permission_checks()
     def test_lookup_filters(self) -> None:
         """Restrict by custom lookup filter."""
         collection = self.playground.create_collection(
@@ -1058,6 +1082,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
                 workflow_root=workflow_root,
             )
 
+    @context.disable_permission_checks()
     def test_alternatives(self) -> None:
         """The list syntax allows specifying several lookups at once."""
         trixie = self.playground.create_collection(
@@ -1093,7 +1118,9 @@ class LookupMultipleTests(LookupMixin, TestCase):
                 ("hello_1.0", {"package": "hello", "version": "1.0"}),
             )
         ]
-        loose_artifacts = [self.create_artifact()[0] for _ in range(3)]
+        loose_artifacts = [
+            self.playground.create_artifact()[0] for _ in range(3)
+        ]
 
         self.assert_lookup_equal(
             [
@@ -1120,9 +1147,10 @@ class LookupMultipleTests(LookupMixin, TestCase):
             expect_type=LookupChildType.ARTIFACT,
         )
 
+    @context.disable_permission_checks()
     def test_artifact_expired(self) -> None:
         """If one artifact has expired, `lookup_multiple` raises LookupError."""
-        artifacts = [self.create_artifact()[0] for _ in range(3)]
+        artifacts = [self.playground.create_artifact()[0] for _ in range(3)]
         artifact_ids = [artifact.id for artifact in artifacts]
         artifacts[0].delete()
 
@@ -1132,6 +1160,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
             expect_type=LookupChildType.ARTIFACT,
         )
 
+    @context.disable_permission_checks()
     def test_collection_removed(self) -> None:
         """If one collection is gone, `lookup_multiple` raises LookupError."""
         collections = [
@@ -1159,6 +1188,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
             f"hidden",
         )
 
+    @context.disable_permission_checks()
     def test_query_count(self) -> None:
         """`lookup_multiple` makes a constant number of DB queries."""
         collection = self.playground.create_collection(
@@ -1204,6 +1234,7 @@ class ReconstructLookupTests(LookupMixin, TestCase):
         item_bare = self.playground.create_bare_data_item(collection, "bare")
         item_artifact = self.create_artifact_item(collection, "artifact")
         item_collection = self.create_collection_item(collection, "collection")
+        workflow = self.playground.create_workflow()
 
         for item, name in (
             (item_bare, "bare"),
@@ -1214,10 +1245,44 @@ class ReconstructLookupTests(LookupMixin, TestCase):
                 reconstruct_lookup(self.make_result(item)),
                 f"{collection.id}@collections/name:{name}",
             )
+            self.assertEqual(
+                reconstruct_lookup(
+                    self.make_result(item), workflow_root=workflow
+                ),
+                f"{collection.id}@collections/name:{name}",
+            )
+
+    def test_collection_item_internal_collection(self) -> None:
+        """Collection items in the internal collection are handled specially."""
+        workflow = self.playground.create_workflow()
+        assert workflow.internal_collection is not None
+        item_bare = self.playground.create_bare_data_item(
+            workflow.internal_collection, "bare"
+        )
+        item_artifact = self.create_artifact_item(
+            workflow.internal_collection, "artifact"
+        )
+        item_collection = self.create_collection_item(
+            workflow.internal_collection, "collection"
+        )
+
+        for item, name in (
+            (item_bare, "bare"),
+            (item_artifact, "artifact"),
+            (item_collection, "collection"),
+        ):
+            self.assertEqual(
+                reconstruct_lookup(
+                    self.make_result(item), workflow_root=workflow
+                ),
+                f"internal@collections/name:{name}",
+            )
 
     def test_artifact(self) -> None:
         """A result with an artifact returns a lookup for that artifact."""
-        artifact, _ = self.create_artifact(category=ArtifactCategory.TEST)
+        artifact, _ = self.playground.create_artifact(
+            category=ArtifactCategory.TEST
+        )
 
         self.assertEqual(
             reconstruct_lookup(self.make_result(artifact)),
@@ -1234,3 +1299,23 @@ class ReconstructLookupTests(LookupMixin, TestCase):
             reconstruct_lookup(self.make_result(collection)),
             f"{collection.id}@collections",
         )
+        self.assertEqual(
+            reconstruct_lookup(
+                self.make_result(collection),
+                workflow_root=self.playground.create_workflow(),
+            ),
+            f"{collection.id}@collections",
+        )
+
+    def test_collection_internal_collection(self) -> None:
+        """The internal collection is handled specially."""
+        workflow = self.playground.create_workflow()
+        assert workflow.internal_collection is not None
+
+        self.assertEqual(
+            reconstruct_lookup(
+                self.make_result(workflow.internal_collection),
+                workflow_root=workflow,
+            ),
+            "internal@collections",
+        )
