@@ -1,4 +1,4 @@
-# Copyright 2024 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -12,9 +12,9 @@
 import io
 
 import yaml
+from django.conf import settings
 from django.core.management import CommandError
 
-from debusine.db.context import context
 from debusine.db.models import WorkRequest, default_workspace, system_user
 from debusine.django.management.tests import call_command
 from debusine.tasks.models import (
@@ -29,7 +29,7 @@ from debusine.test.django import TestCase
 class CreateWorkRequestCommandTests(TestCase):
     """Tests for the create_work_request command."""
 
-    def test_data_from_file(self):
+    def test_data_from_file(self) -> None:
         """`create_work_request` accepts data from a file."""
         data = {"result": True}
         data_file = self.create_temporary_file(
@@ -56,7 +56,7 @@ class CreateWorkRequestCommandTests(TestCase):
         self.assertEqual(stderr, "")
         self.assertEqual(exit_code, 0)
 
-    def test_data_from_stdin(self):
+    def test_data_from_stdin(self) -> None:
         """`create_work_request` accepts data from stdin."""
         data = {"result": True}
         stdout, stderr, exit_code = call_command(
@@ -79,7 +79,7 @@ class CreateWorkRequestCommandTests(TestCase):
         self.assertEqual(stderr, "")
         self.assertEqual(exit_code, 0)
 
-    def test_event_reactions(self):
+    def test_event_reactions(self) -> None:
         """`create_work_request` handles event reactions."""
         data = {"result": True}
         event_reactions = {
@@ -113,7 +113,7 @@ class CreateWorkRequestCommandTests(TestCase):
         self.assertEqual(stderr, "")
         self.assertEqual(exit_code, 0)
 
-    def test_server(self):
+    def test_server(self) -> None:
         """`create_work_request` can create server work requests."""
         data = {"result": True}
         event_reactions = {
@@ -159,7 +159,7 @@ class CreateWorkRequestCommandTests(TestCase):
         self.assertEqual(stderr, "")
         self.assertEqual(exit_code, 0)
 
-    def test_signing(self):
+    def test_signing(self) -> None:
         """`create_work_request` can create signing work requests."""
         data = {"purpose": "uefi", "description": "A UEFI key"}
         event_reactions = {
@@ -205,7 +205,7 @@ class CreateWorkRequestCommandTests(TestCase):
         self.assertEqual(stderr, "")
         self.assertEqual(exit_code, 0)
 
-    def test_empty_data(self):
+    def test_empty_data(self) -> None:
         """`create_work_request` defaults data to {}."""
         stdout, stderr, exit_code = call_command(
             "create_work_request", "worker", "noop", stdin=io.StringIO()
@@ -221,7 +221,7 @@ class CreateWorkRequestCommandTests(TestCase):
         self.assertEqual(stderr, "")
         self.assertEqual(exit_code, 0)
 
-    def test_different_created_by(self):
+    def test_different_created_by(self) -> None:
         """`create_work_request` can use a different created-by user."""
         user = self.playground.get_default_user()
         stdout, stderr, exit_code = call_command(
@@ -243,11 +243,10 @@ class CreateWorkRequestCommandTests(TestCase):
         self.assertEqual(stderr, "")
         self.assertEqual(exit_code, 0)
 
-    def test_different_workspace(self):
+    def test_different_workspace(self) -> None:
         """`create_work_request` can use a non-default workspace."""
         workspace_name = "test-workspace"
-        with context.disable_permission_checks():
-            self.playground.create_workspace(name=workspace_name)
+        workspace = self.playground.create_workspace(name=workspace_name)
         stdout, stderr, exit_code = call_command(
             "create_work_request",
             "worker",
@@ -260,14 +259,40 @@ class CreateWorkRequestCommandTests(TestCase):
         work_request = WorkRequest.objects.get(
             task_type=TaskTypes.WORKER, task_name="noop"
         )
-        self.assertEqual(work_request.workspace.name, workspace_name)
+        self.assertEqual(work_request.workspace, workspace)
 
         expected = yaml.safe_dump({"work_request_id": work_request.id})
         self.assertEqual(stdout, f"---\n{expected}\n")
         self.assertEqual(stderr, "")
         self.assertEqual(exit_code, 0)
 
-    def test_invalid_data_yaml(self):
+    def test_different_scope(self) -> None:
+        """`create_work_request` can use a workspace in a different scope."""
+        scope_name = "test-scope"
+        scope = self.playground.get_or_create_scope(name=scope_name)
+        workspace = self.playground.create_workspace(
+            scope=scope, name=settings.DEBUSINE_DEFAULT_WORKSPACE
+        )
+        stdout, stderr, exit_code = call_command(
+            "create_work_request",
+            "worker",
+            "noop",
+            "--workspace",
+            f"{scope_name}/{settings.DEBUSINE_DEFAULT_WORKSPACE}",
+            stdin=io.StringIO(),
+        )
+
+        work_request = WorkRequest.objects.get(
+            task_type=TaskTypes.WORKER, task_name="noop"
+        )
+        self.assertEqual(work_request.workspace, workspace)
+
+        expected = yaml.safe_dump({"work_request_id": work_request.id})
+        self.assertEqual(stdout, f"---\n{expected}\n")
+        self.assertEqual(stderr, "")
+        self.assertEqual(exit_code, 0)
+
+    def test_invalid_data_yaml(self) -> None:
         """`create_work_request` returns error: cannot parse data."""
         with self.assertRaisesRegex(
             CommandError, r"^Error parsing YAML:"
@@ -278,7 +303,7 @@ class CreateWorkRequestCommandTests(TestCase):
 
         self.assertEqual(exc.exception.returncode, 3)
 
-    def test_invalid_event_reactions_yaml(self):
+    def test_invalid_event_reactions_yaml(self) -> None:
         """`create_work_request` returns error: cannot parse data."""
         event_reactions_file = self.create_temporary_file(contents=b":")
 
@@ -296,7 +321,7 @@ class CreateWorkRequestCommandTests(TestCase):
 
         self.assertEqual(exc.exception.returncode, 3)
 
-    def test_bad_task_type(self):
+    def test_bad_task_type(self) -> None:
         """`create_work_request` returns error: bad task type."""
         with self.assertRaisesRegex(
             CommandError, r"task_type: invalid choice: 'nonexistent'"
@@ -310,7 +335,7 @@ class CreateWorkRequestCommandTests(TestCase):
 
         self.assertEqual(exc.exception.returncode, 1)
 
-    def test_user_not_found(self):
+    def test_user_not_found(self) -> None:
         """`create_work_request` returns error: user not found."""
         with self.assertRaisesRegex(
             CommandError, r'^User "nonexistent" not found'
@@ -326,10 +351,10 @@ class CreateWorkRequestCommandTests(TestCase):
 
         self.assertEqual(exc.exception.returncode, 3)
 
-    def test_workspace_not_found(self):
+    def test_workspace_not_found(self) -> None:
         """`create_work_request` returns error: workspace not found."""
         with self.assertRaisesRegex(
-            CommandError, r'^Workspace "nonexistent" not found'
+            CommandError, r"^Workspace 'nonexistent' not found"
         ) as exc:
             call_command(
                 "create_work_request",
@@ -342,7 +367,7 @@ class CreateWorkRequestCommandTests(TestCase):
 
         self.assertEqual(exc.exception.returncode, 3)
 
-    def test_task_name_not_found(self):
+    def test_task_name_not_found(self) -> None:
         """`create_work_request` returns error: task name not found."""
         with self.assertRaisesRegex(
             CommandError,
@@ -357,7 +382,7 @@ class CreateWorkRequestCommandTests(TestCase):
 
         self.assertEqual(exc.exception.returncode, 3)
 
-    def test_bad_task_data(self):
+    def test_bad_task_data(self) -> None:
         """`create_work_request` returns error: bad task data."""
         with self.assertRaisesRegex(
             CommandError, r"invalid Worker task data"
