@@ -70,7 +70,6 @@ class MakeSignedSourceWorkflow(
             configuration_key="binary_artifacts",
             artifact_expected_categories=(
                 ArtifactCategory.BINARY_PACKAGE,
-                ArtifactCategory.BINARY_PACKAGES,
                 ArtifactCategory.UPLOAD,
             ),
         )
@@ -138,6 +137,7 @@ class MakeSignedSourceWorkflow(
                 sbuild_architectures = sorted(architectures | {"all"})
                 self._populate_sbuild_workflow(
                     assembled=assembled,
+                    extra_binary_artifacts=binary_artifacts_lookups,
                     target_distribution=target_distribution,
                     backend=self.data.sbuild_backend,
                     architectures=sbuild_architectures,
@@ -290,6 +290,7 @@ class MakeSignedSourceWorkflow(
     def _populate_sbuild_workflow(
         self,
         assembled: LookupSingle,
+        extra_binary_artifacts: LookupMultiple,
         target_distribution: str,
         backend: BackendType,
         architectures: list[str],
@@ -301,7 +302,10 @@ class MakeSignedSourceWorkflow(
             task_name="sbuild",
             task_data=SbuildWorkflowData(
                 prefix="signed-source/",
-                input=SbuildInput(source_artifact=assembled),
+                input=SbuildInput(
+                    source_artifact=assembled,
+                    extra_binary_artifacts=extra_binary_artifacts,
+                ),
                 target_distribution=target_distribution,
                 backend=backend,
                 architectures=architectures,
