@@ -1,4 +1,4 @@
-# Copyright 2024 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -8,27 +8,43 @@
 # contained in the LICENSE file.
 
 """Unit tests for the debian pipeline workflow."""
-from typing import Any
-from unittest.mock import call, patch
+from typing import Any, ClassVar
 
-from debusine.artifacts.models import (
-    ArtifactCategory,
-    CollectionCategory,
-    KeyPurpose,
-)
-from debusine.db.models import WorkRequest, WorkflowTemplate
-from debusine.server.workflows import DebianPipelineWorkflow
+from debusine.artifacts.models import ArtifactCategory, CollectionCategory
+from debusine.assets import KeyPurpose
+from debusine.client.models import LookupChildType
+from debusine.db.context import context
+from debusine.db.models import Artifact, TaskDatabase, WorkRequest
+from debusine.server.workflows import DebianPipelineWorkflow, workflow_utils
 from debusine.server.workflows.base import (
     WorkflowRunError,
     orchestrate_workflow,
 )
-from debusine.tasks.models import BackendType, LookupMultiple, TaskTypes
+from debusine.server.workflows.models import DebianPipelineWorkflowData
+from debusine.tasks import TaskConfigError
+from debusine.tasks.models import (
+    BackendType,
+    BaseDynamicTaskData,
+    CollectionItemMatcherKind,
+    TaskTypes,
+)
 from debusine.test.django import TestCase
 
 
 class DebianPipelineWorkflowTests(TestCase):
     """Unit tests for :py:class:`DebianPipelineWorkflow`."""
 
+    source_artifact: ClassVar[Artifact]
+
+    @classmethod
+    @context.disable_permission_checks()
+    def setUpTestData(cls) -> None:
+        """Set up common data."""
+        super().setUpTestData()
+        cls.source_artifact = cls.playground.create_source_artifact(
+            name="hello"
+        )
+
     def create_debian_pipeline_workflow(
         self, *, extra_task_data: dict[str, Any]
     ) -> DebianPipelineWorkflow:
@@ -46,12 +62,6 @@ class DebianPipelineWorkflowTests(TestCase):
 
     def orchestrate(self, extra_data: dict[str, Any]) -> WorkRequest:
         """Create and orchestrate a DebianPipelineWorkflow."""
-        template = WorkflowTemplate(
-            name="debianpipeline",
-            workspace=self.playground.get_default_workspace(),
-            task_name="debian_pipeline",
-        )
-
         data = {
             "source_artifact": 10,
             "vendor": "debian",
@@ -59,14 +69,12 @@ class DebianPipelineWorkflowTests(TestCase):
             "enable_autopkgtest": False,
             "enable_lintian": False,
             "enable_piuparts": False,
+            **extra_data,
         }
 
-        data.update(extra_data)
-
-        wr = WorkRequest.objects.create_workflow(
-            template=template,
-            data=data,
-            created_by=self.playground.get_default_user(),
+        wr = self.playground.create_workflow(
+            task_name="debian_pipeline",
+            task_data=DebianPipelineWorkflowData(**data),
         )
         self.assertEqual(wr.status, WorkRequest.Statuses.PENDING)
         wr.mark_running()
@@ -74,32 +82,89 @@ class DebianPipelineWorkflowTests(TestCase):
 
         return wr
 
-    def test_populate_use_available_architectures(self):
+    def test_compute_dynamic_data(self) -> None:
+        """Dynamic data includes a parameter summary."""
+        source_artifact = self.playground.create_source_artifact(
+            name="hello", version="1.0-1"
+        )
+        wr = self.playground.create_workflow(
+            task_name="debian_pipeline",
+            task_data=DebianPipelineWorkflowData(
+                source_artifact=source_artifact.id,
+                vendor="debian",
+                codename="sid",
+            ),
+        )
+        workflow = DebianPipelineWorkflow(wr)
+
+        self.assertEqual(
+            workflow.compute_dynamic_data(TaskDatabase(wr)),
+            BaseDynamicTaskData(
+                subject="hello", parameter_summary="hello_1.0-1"
+            ),
+        )
+
+    def test_compute_dynamic_data_source_artifact_wrong_category(self) -> None:
+        """`source_artifact` must be a source-package or upload."""
+        artifact, _ = self.playground.create_artifact(
+            category=ArtifactCategory.TEST
+        )
+        wr = self.playground.create_workflow(
+            task_name="debian_pipeline",
+            task_data=DebianPipelineWorkflowData(
+                source_artifact=artifact.id, vendor="debian", codename="sid"
+            ),
+        )
+        workflow = DebianPipelineWorkflow(wr)
+
+        with self.assertRaisesRegex(
+            TaskConfigError,
+            r"^source_artifact: unexpected artifact category: 'debusine:test'. "
+            r"Valid categories: \['debian:source-package', 'debian:upload'\]$",
+        ):
+            workflow.compute_dynamic_data(TaskDatabase(wr))
+
+    def test_source_package_upload(self) -> None:
+        """`source_artifact` can be an upload."""
+        upload_artifacts = self.playground.create_upload_artifacts()
+        wr = self.playground.create_workflow(
+            task_name="debian_pipeline",
+            task_data=DebianPipelineWorkflowData(
+                source_artifact=upload_artifacts.upload.id,
+                vendor="debian",
+                codename="sid",
+            ),
+        )
+        workflow = DebianPipelineWorkflow(wr)
+        self.assertEqual(
+            workflow_utils.source_package(workflow), upload_artifacts.source
+        )
+
+    def test_populate_use_available_architectures(self) -> None:
         """
         Test populate use available architectures.
 
         The user didn't specify "architectures", DebianPipelineWorkflow
         checks available architectures and "all" and use them.
         """
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
-        collection = self.playground.create_collection(
-            "debian",
-            CollectionCategory.ENVIRONMENTS,
-            workspace=self.playground.get_default_workspace(),
-        )
-        for arch in ["amd64", "i386"]:
-            artifact, _ = self.create_artifact(
-                category=ArtifactCategory.SYSTEM_TARBALL,
-                data={"codename": "bookworm", "architecture": arch},
-            )
-            collection.manager.add_artifact(
-                artifact, user=self.playground.get_default_user()
+        with context.disable_permission_checks():
+            collection = self.playground.create_collection(
+                "debian",
+                CollectionCategory.ENVIRONMENTS,
+                workspace=self.playground.get_default_workspace(),
             )
+            for arch in ["amd64", "i386"]:
+                artifact, _ = self.playground.create_artifact(
+                    category=ArtifactCategory.SYSTEM_TARBALL,
+                    data={"codename": "bookworm", "architecture": arch},
+                )
+                collection.manager.add_artifact(
+                    artifact, user=self.playground.get_default_user()
+                )
 
         workflow = self.orchestrate(
             extra_data={
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "source_artifact": self.source_artifact.id,
                 "vendor": "debian",
                 "codename": "bookworm",
             }
@@ -113,14 +178,12 @@ class DebianPipelineWorkflowTests(TestCase):
             sbuild.task_data["architectures"], ["all", "amd64", "i386"]
         )
 
-    def test_populate_sbuild(self):
+    def test_populate_sbuild(self) -> None:
         """Test populate create sbuild."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
         workflow = self.orchestrate(
             extra_data={
                 "architectures": ["amd64"],
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "source_artifact": self.source_artifact.id,
             }
         )
 
@@ -128,13 +191,16 @@ class DebianPipelineWorkflowTests(TestCase):
             task_name="sbuild", task_type=TaskTypes.WORKFLOW
         )
 
+        self.assertEqual(sbuild.status, WorkRequest.Statuses.RUNNING)
         self.assertEqual(
             sbuild.task_data,
             {
+                "arch_all_host_architecture": "amd64",
                 "architectures": ["amd64"],
                 "backend": BackendType.AUTO,
                 "environment_variant": None,
-                "input": {"source_artifact": f"{source_artifact.id}@artifacts"},
+                "extra_repositories": None,
+                "input": {"source_artifact": self.source_artifact.id},
                 "target_distribution": "debian:bookworm",
                 "signing_template_names": {},
             },
@@ -146,6 +212,8 @@ class DebianPipelineWorkflowTests(TestCase):
 
         # SbuildWorkflow.populate() was called and created its tasks
         self.assertTrue(sbuild.children.exists())
+        for child in sbuild.children.all():
+            self.assertEqual(child.status, WorkRequest.Statuses.PENDING)
 
         self.assertFalse(
             workflow.children.filter(
@@ -158,32 +226,118 @@ class DebianPipelineWorkflowTests(TestCase):
     ) -> None:
         """Assert workflow has a sub-workflow qa with task_data."""
         qa = workflow.children.get(task_name="qa", task_type=TaskTypes.WORKFLOW)
+        self.assertEqual(qa.status, WorkRequest.Statuses.RUNNING)
         self.assertEqual(qa.task_data, task_data)
 
         self.assertEqual(
             qa.workflow_data_json, {"display_name": "QA", "step": "qa"}
         )
 
-        self.assertQuerySetEqual(
-            qa.dependencies.all(),
-            list(
-                WorkRequest.objects.filter(
-                    task_type=TaskTypes.WORKER, task_name="sbuild"
-                )
-            ),
-        )
-
         # QaWorkflow.populate() was called and created its tasks
         self.assertTrue(qa.children.exists())
+        for child in qa.children.all():
+            # Children are sub-workflows, so go straight to RUNNING.
+            self.assertEqual(child.status, WorkRequest.Statuses.RUNNING)
 
-    def test_populate_qa_lintian(self):
+        self.assertQuerySetEqual(
+            qa.children.earliest("id").dependencies.all(), []
+        )
+
+    def test_populate_qa_with_sbuild_subset(self) -> None:
+        """`sbuild` may only create a subset of the requested architectures."""
+        self.assertEqual(
+            self.source_artifact.data["dsc_fields"]["Architecture"], "any"
+        )
+
+        workflow = self.orchestrate(
+            extra_data={
+                "architectures": ["all", "amd64"],
+                "source_artifact": f"{self.source_artifact.id}@artifacts",
+                "enable_lintian": True,
+                "extra_repositories": [
+                    {
+                        "url": "http://example.com/",
+                        "suite": "bookworm",
+                        "components": ["main"],
+                    }
+                ],
+            }
+        )
+
+        sbuild = workflow.children.get(
+            task_name="sbuild", task_type=TaskTypes.WORKFLOW
+        )
+
+        self.assertEqual(
+            sbuild.task_data,
+            {
+                "arch_all_host_architecture": "amd64",
+                "architectures": ["all", "amd64"],
+                "backend": BackendType.AUTO,
+                "environment_variant": None,
+                "input": {
+                    "source_artifact": f"{self.source_artifact.id}@artifacts"
+                },
+                "target_distribution": "debian:bookworm",
+                "signing_template_names": {},
+                "extra_repositories": [
+                    {
+                        "url": "http://example.com/",
+                        "suite": "bookworm",
+                        "components": ["main"],
+                    }
+                ],
+            },
+        )
+        self.assertEqual(
+            sbuild.workflow_data_json,
+            {"display_name": "sbuild", "step": "sbuild"},
+        )
+
+        self.assertTrue(sbuild.children.exists())
+        self.assert_qa(
+            workflow,
+            {
+                "arch_all_host_architecture": "amd64",
+                "architectures": ["all", "amd64"],
+                "autopkgtest_backend": BackendType.AUTO,
+                "binary_artifacts": [
+                    {
+                        "collection": "internal@collections",
+                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
+                        "name_matcher": {
+                            "kind": CollectionItemMatcherKind.STARTSWITH,
+                            "value": "build-",
+                        },
+                    }
+                ],
+                "codename": "bookworm",
+                "extra_repositories": [
+                    {
+                        "url": "http://example.com/",
+                        "suite": "bookworm",
+                        "components": ["main"],
+                    }
+                ],
+                "enable_autopkgtest": False,
+                "enable_lintian": True,
+                "enable_piuparts": False,
+                "enable_reverse_dependencies_autopkgtest": False,
+                "lintian_backend": BackendType.AUTO,
+                "lintian_fail_on_severity": "none",
+                "piuparts_backend": BackendType.AUTO,
+                "reverse_dependencies_autopkgtest_suite": None,
+                "source_artifact": f"{self.source_artifact.id}@artifacts",
+                "vendor": "debian",
+            },
+        )
+
+    def test_populate_qa_lintian(self) -> None:
         """Test populate create qa: lintian enabled."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
         workflow = self.orchestrate(
             extra_data={
                 "architectures": ["amd64"],
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "source_artifact": self.source_artifact.id,
                 "enable_lintian": True,
             }
         )
@@ -193,27 +347,37 @@ class DebianPipelineWorkflowTests(TestCase):
                 "arch_all_host_architecture": "amd64",
                 "architectures": ["amd64"],
                 "autopkgtest_backend": BackendType.AUTO,
-                "binary_artifacts": ["internal@collections/name:build-amd64"],
+                "binary_artifacts": [
+                    {
+                        "collection": "internal@collections",
+                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
+                        "name_matcher": {
+                            "kind": CollectionItemMatcherKind.STARTSWITH,
+                            "value": "build-",
+                        },
+                    }
+                ],
                 "codename": "bookworm",
+                "extra_repositories": None,
                 "enable_autopkgtest": False,
                 "enable_lintian": True,
                 "enable_piuparts": False,
+                "enable_reverse_dependencies_autopkgtest": False,
                 "lintian_backend": BackendType.AUTO,
                 "lintian_fail_on_severity": "none",
                 "piuparts_backend": BackendType.AUTO,
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "reverse_dependencies_autopkgtest_suite": None,
+                "source_artifact": self.source_artifact.id,
                 "vendor": "debian",
             },
         )
 
-    def test_populate_qa_autopkgtest(self):
+    def test_populate_qa_autopkgtest(self) -> None:
         """Test populate create qa: autopkgtest enabled."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
         workflow = self.orchestrate(
             extra_data={
                 "architectures": ["amd64"],
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "source_artifact": self.source_artifact.id,
                 "enable_autopkgtest": True,
             }
         )
@@ -224,27 +388,37 @@ class DebianPipelineWorkflowTests(TestCase):
                 "arch_all_host_architecture": "amd64",
                 "architectures": ["amd64"],
                 "autopkgtest_backend": BackendType.AUTO,
-                "binary_artifacts": ["internal@collections/name:build-amd64"],
+                "binary_artifacts": [
+                    {
+                        "collection": "internal@collections",
+                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
+                        "name_matcher": {
+                            "kind": CollectionItemMatcherKind.STARTSWITH,
+                            "value": "build-",
+                        },
+                    }
+                ],
                 "codename": "bookworm",
+                "extra_repositories": None,
                 "enable_autopkgtest": True,
                 "enable_lintian": False,
                 "enable_piuparts": False,
+                "enable_reverse_dependencies_autopkgtest": False,
                 "lintian_backend": BackendType.AUTO,
                 "lintian_fail_on_severity": "none",
                 "piuparts_backend": BackendType.AUTO,
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "reverse_dependencies_autopkgtest_suite": None,
+                "source_artifact": self.source_artifact.id,
                 "vendor": "debian",
             },
         )
 
-    def test_populate_qa_piuparts(self):
+    def test_populate_qa_piuparts(self) -> None:
         """Test populate create qa: piuparts enabled."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
         workflow = self.orchestrate(
             extra_data={
                 "architectures": ["amd64"],
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "source_artifact": self.source_artifact.id,
                 "enable_piuparts": True,
             }
         )
@@ -255,26 +429,36 @@ class DebianPipelineWorkflowTests(TestCase):
                 "arch_all_host_architecture": "amd64",
                 "architectures": ["amd64"],
                 "autopkgtest_backend": BackendType.AUTO,
-                "binary_artifacts": ["internal@collections/name:build-amd64"],
+                "binary_artifacts": [
+                    {
+                        "collection": "internal@collections",
+                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
+                        "name_matcher": {
+                            "kind": CollectionItemMatcherKind.STARTSWITH,
+                            "value": "build-",
+                        },
+                    }
+                ],
                 "codename": "bookworm",
+                "extra_repositories": None,
                 "enable_autopkgtest": False,
                 "enable_lintian": False,
                 "enable_piuparts": True,
+                "enable_reverse_dependencies_autopkgtest": False,
                 "lintian_backend": BackendType.AUTO,
                 "lintian_fail_on_severity": "none",
                 "piuparts_backend": BackendType.AUTO,
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "reverse_dependencies_autopkgtest_suite": None,
+                "source_artifact": self.source_artifact.id,
                 "vendor": "debian",
             },
         )
 
-    def test_populate_architectures_allow_deny(self):
+    def test_populate_architectures_allow_deny(self) -> None:
         """Populate uses architectures, architectures_allow/deny."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
         workflow = self.orchestrate(
             extra_data={
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "source_artifact": self.source_artifact.id,
                 "architectures": ["amd64", "i386", "arm64"],
                 "architectures_allowlist": ["amd64", "i386"],
                 "architectures_denylist": ["amd64"],
@@ -287,19 +471,21 @@ class DebianPipelineWorkflowTests(TestCase):
         self.assertEqual(
             sbuild.task_data,
             {
+                "arch_all_host_architecture": "amd64",
                 "architectures": ["i386"],
                 "backend": BackendType.AUTO,
                 "environment_variant": None,
-                "input": {"source_artifact": f"{source_artifact.id}@artifacts"},
+                "extra_repositories": None,
+                "input": {"source_artifact": self.source_artifact.id},
                 "target_distribution": "debian:bookworm",
                 "signing_template_names": {},
             },
         )
 
-    def test_populate_make_signed_source_missing_signed_source_purpose(self):
+    def test_populate_make_signed_source_missing_signed_source_purpose(
+        self,
+    ) -> None:
         """Populate raise WorkflowRunError: missing required field."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
         msg = (
             'orchestrator failed: "make_signed_source_purpose" must be '
             'set when signing the source'
@@ -307,17 +493,17 @@ class DebianPipelineWorkflowTests(TestCase):
         with self.assertRaisesRegex(WorkflowRunError, msg):
             self.orchestrate(
                 extra_data={
-                    "source_artifact": f"{source_artifact.id}@artifacts",
+                    "source_artifact": self.source_artifact.id,
                     "architectures": ["amd64"],
                     "enable_make_signed_source": True,
                     "signing_template_names": {"amd64": ["hello"]},
                 }
             )
 
-    def test_populate_make_signed_source_missing_make_signed_source_key(self):
+    def test_populate_make_signed_source_missing_make_signed_source_key(
+        self,
+    ) -> None:
         """Populate raise WorkflowRunError: missing required field."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
         msg = (
             'orchestrator failed: "make_signed_source_key" must be set '
             'when signing the source'
@@ -325,7 +511,7 @@ class DebianPipelineWorkflowTests(TestCase):
         with self.assertRaisesRegex(WorkflowRunError, msg):
             self.orchestrate(
                 extra_data={
-                    "source_artifact": f"{source_artifact.id}@artifacts",
+                    "source_artifact": self.source_artifact.id,
                     "architectures": ["amd64"],
                     "enable_make_signed_source": True,
                     "make_signed_source_purpose": KeyPurpose.OPENPGP,
@@ -333,28 +519,29 @@ class DebianPipelineWorkflowTests(TestCase):
                 }
             )
 
-    def test_populate_make_signed_source(self):
+    def test_populate_make_signed_source(self) -> None:
         """Populate create a make_signed_source workflow."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
-        with patch.object(
-            DebianPipelineWorkflow,
-            "requires_artifact",
-            wraps=DebianPipelineWorkflow.requires_artifact,
-        ) as requires_artifact_wrapper:
-            workflow = self.orchestrate(
-                extra_data={
-                    "architectures": ["amd64"],
-                    "source_artifact": f"{source_artifact.id}@artifacts",
-                    "enable_make_signed_source": True,
-                    "make_signed_source_purpose": KeyPurpose.OPENPGP,
-                    "signing_template_names": {
-                        "amd64": ["hello"],
-                    },
-                    "make_signed_source_key": 80,
-                }
-            )
+        workflow = self.orchestrate(
+            extra_data={
+                "architectures": ["amd64"],
+                "source_artifact": self.source_artifact.id,
+                "enable_make_signed_source": True,
+                "make_signed_source_purpose": KeyPurpose.OPENPGP,
+                "signing_template_names": {"amd64": ["hello"]},
+                "make_signed_source_key": "ABC123",
+            }
+        )
 
+        expected_binary_artifacts = [
+            {
+                "collection": "internal@collections",
+                "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
+                "name_matcher": {
+                    "kind": CollectionItemMatcherKind.STARTSWITH,
+                    "value": "build-",
+                },
+            }
+        ]
         make_signed_source = workflow.children.get(
             task_name="make_signed_source", task_type=TaskTypes.WORKFLOW
         )
@@ -362,11 +549,9 @@ class DebianPipelineWorkflowTests(TestCase):
             make_signed_source.task_data,
             {
                 "architectures": ["amd64"],
-                "binary_artifacts": [
-                    "internal@collections/name:build-amd64",
-                ],
+                "binary_artifacts": expected_binary_artifacts,
                 "codename": "bookworm",
-                "key": 80,
+                "key": "ABC123",
                 "purpose": "openpgp",
                 "sbuild_backend": BackendType.AUTO,
                 "signing_template_artifacts": [
@@ -383,58 +568,61 @@ class DebianPipelineWorkflowTests(TestCase):
                 "step": "make_signed_source",
             },
         )
+        self.assertQuerySetEqual(make_signed_source.dependencies.all(), [])
 
-        # There are two calls to self.requires_artifact:
-        # self.requires_artifact(wr, binary_artifacts)
-        # self.requires_artifact(wr, LookupMultiple.parse_obj...)
-        # But both of them are adding the same WorkRequest to depend on:
-        # the sbuild work request (which generates the binary artifacts and the
-        # signing artifacts
+        # MakeSignedSource.populate() was called and created its tasks.
+        # They are all blocked, with the first one depending on sbuild and
+        # the others depending on the previous one in sequence.
         self.assertQuerySetEqual(
-            make_signed_source.dependencies.all(),
-            list(
-                WorkRequest.objects.filter(
-                    task_type=TaskTypes.WORKER, task_name="sbuild"
-                )
+            make_signed_source.children.order_by("id").values_list(
+                "task_type",
+                "task_name",
+                "status",
+                "dependencies__task_type",
+                "dependencies__task_name",
             ),
-        )
-
-        # Assert that the correct calls were made
-        requires_artifact_wrapper.assert_has_calls(
             [
-                call(
-                    make_signed_source,
-                    LookupMultiple.parse_obj(
-                        ["internal@collections/name:build-amd64"]
-                    ),
+                (
+                    TaskTypes.WORKER,
+                    "extractforsigning",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WORKER,
+                    "sbuild",
                 ),
-                call(
-                    make_signed_source,
-                    LookupMultiple.parse_obj(
-                        [
-                            "internal@collections/"
-                            "name:signing-template-amd64-hello"
-                        ]
-                    ),
+                (
+                    TaskTypes.SIGNING,
+                    "sign",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WORKER,
+                    "extractforsigning",
                 ),
-            ]
+                (
+                    TaskTypes.WORKER,
+                    "assemblesignedsource",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.SIGNING,
+                    "sign",
+                ),
+                (
+                    TaskTypes.WORKFLOW,
+                    "sbuild",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WORKER,
+                    "assemblesignedsource",
+                ),
+            ],
         )
 
-        # MakeSignedSource.populate() was called and created its tasks
-        self.assertTrue(make_signed_source.children.exists())
-
-    def test_populate_package_upload(self):
+    def test_populate_package_upload(self) -> None:
         """Populate create an upload_package workflow."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
         workflow = self.orchestrate(
             extra_data={
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "source_artifact": self.source_artifact.id,
                 "architectures": ["amd64", "i386"],
                 "enable_make_signed_source": True,
                 "make_signed_source_purpose": KeyPurpose.OPENPGP,
                 "signing_template_names": {"amd64": ["hello"]},
-                "make_signed_source_key": 80,
+                "make_signed_source_key": "ABC123",
                 "enable_upload": True,
                 "vendor": "debian",
                 "codename": "trixie",
@@ -453,15 +641,22 @@ class DebianPipelineWorkflowTests(TestCase):
             package_upload.task_data,
             {
                 "binary_artifacts": [
-                    "internal@collections/name:build-amd64",
-                    "internal@collections/name:build-i386",
+                    {
+                        "collection": "internal@collections",
+                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
+                        "name_matcher": {
+                            "kind": CollectionItemMatcherKind.STARTSWITH,
+                            "value": "build-",
+                        },
+                    }
                 ],
                 "codename": "trixie",
+                "delayed_days": None,
                 "key": None,
                 "merge_uploads": False,
                 "require_signature": True,
                 "since_version": "1.1",
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "source_artifact": self.source_artifact.id,
                 "target": "ftp://user@example.org/pub/UploadQueue1/",
                 "target_distribution": "debian:testing",
                 "vendor": "debian",
@@ -472,31 +667,73 @@ class DebianPipelineWorkflowTests(TestCase):
             package_upload.workflow_data_json,
             {"display_name": "package upload", "step": "package_upload"},
         )
+        self.assertQuerySetEqual(package_upload.dependencies.all(), [])
 
+        # PackageUpload.populate() was called and created its tasks.
+        # MakeSourcePackageUpload only requires the source artifact and so
+        # can start immediately.  Everything else is blocked.
         self.assertQuerySetEqual(
-            package_upload.dependencies.all(),
-            list(
-                WorkRequest.objects.filter(
-                    task_type=TaskTypes.WORKER, task_name="sbuild"
-                )
+            package_upload.children.order_by("id").values_list(
+                "task_type",
+                "task_name",
+                "status",
+                "dependencies__task_type",
+                "dependencies__task_name",
             ),
+            # Work requests to construct, sign, and upload the source package.
+            [
+                (
+                    TaskTypes.WORKER,
+                    "makesourcepackageupload",
+                    WorkRequest.Statuses.PENDING,
+                    None,
+                    None,
+                ),
+                (
+                    TaskTypes.WAIT,
+                    "externaldebsign",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WORKER,
+                    "makesourcepackageupload",
+                ),
+                (
+                    TaskTypes.SERVER,
+                    "packageupload",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WAIT,
+                    "externaldebsign",
+                ),
+            ]
+            # Work requests to sign and upload each of the two binary packages.
+            + [
+                (
+                    TaskTypes.WAIT,
+                    "externaldebsign",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WORKER,
+                    "sbuild",
+                ),
+                (
+                    TaskTypes.SERVER,
+                    "packageupload",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WAIT,
+                    "externaldebsign",
+                ),
+            ]
+            * 2,
         )
 
-        # PackageUpload.populate() was called and created its tasks
-        self.assertTrue(package_upload.children.exists())
-
-    def test_populate_upload_package_no_include_source_binaries(self):
+    def test_populate_upload_package_no_include_source_binaries(self) -> None:
         """Populate create upload_workflow without source neither binaries."""
-        source_artifact = self.playground.create_source_artifact(name="hello")
-
         workflow = self.orchestrate(
             extra_data={
                 "architectures": ["amd64"],
-                "source_artifact": f"{source_artifact.id}@artifacts",
+                "source_artifact": self.source_artifact.id,
                 "enable_make_signed_source": True,
                 "make_signed_source_purpose": KeyPurpose.OPENPGP,
                 "signing_template_names": {"amd64": ["hello"]},
-                "make_signed_source_key": 80,
+                "make_signed_source_key": "ABC123",
                 "enable_upload": True,
                 "vendor": "debian",
                 "codename": "trixie",
@@ -517,6 +754,7 @@ class DebianPipelineWorkflowTests(TestCase):
             {
                 "binary_artifacts": [],
                 "codename": "trixie",
+                "delayed_days": None,
                 "key": None,
                 "merge_uploads": False,
                 "require_signature": True,
@@ -533,7 +771,113 @@ class DebianPipelineWorkflowTests(TestCase):
             {"display_name": "package upload", "step": "package_upload"},
         )
 
-    def test_get_label(self):
+    def test_populate_package_upload_delayed(self) -> None:
+        """Populate an upload_package workflow for a delayed upload."""
+        workflow = self.orchestrate(
+            extra_data={
+                "source_artifact": self.source_artifact.id,
+                "architectures": ["amd64", "i386"],
+                "enable_upload": True,
+                "vendor": "debian",
+                "codename": "trixie",
+                "upload_merge_uploads": False,
+                "upload_since_version": "1.1",
+                "upload_target": "ftp://user@example.org/pub/UploadQueue1/",
+                "upload_target_distribution": "debian:testing",
+                "upload_delayed_days": 3,
+            }
+        )
+
+        package_upload = workflow.children.get(
+            task_name="package_upload", task_type=TaskTypes.WORKFLOW
+        )
+
+        self.assertEqual(
+            package_upload.task_data,
+            {
+                "binary_artifacts": [
+                    {
+                        "collection": "internal@collections",
+                        "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
+                        "name_matcher": {
+                            "kind": CollectionItemMatcherKind.STARTSWITH,
+                            "value": "build-",
+                        },
+                    }
+                ],
+                "codename": "trixie",
+                "delayed_days": 3,
+                "key": None,
+                "merge_uploads": False,
+                "require_signature": True,
+                "since_version": "1.1",
+                "source_artifact": self.source_artifact.id,
+                "target": "ftp://user@example.org/pub/UploadQueue1/",
+                "target_distribution": "debian:testing",
+                "vendor": "debian",
+            },
+        )
+
+        self.assertEqual(
+            package_upload.workflow_data_json,
+            {"display_name": "package upload", "step": "package_upload"},
+        )
+        self.assertQuerySetEqual(package_upload.dependencies.all(), [])
+
+        # PackageUpload.populate() was called and created its tasks.
+        # MakeSourcePackageUpload only requires the source artifact and so
+        # can start immediately.  Everything else is blocked.
+        self.assertQuerySetEqual(
+            package_upload.children.order_by("id").values_list(
+                "task_type",
+                "task_name",
+                "status",
+                "dependencies__task_type",
+                "dependencies__task_name",
+            ),
+            [
+                (
+                    TaskTypes.WORKER,
+                    "makesourcepackageupload",
+                    WorkRequest.Statuses.PENDING,
+                    None,
+                    None,
+                ),
+                (
+                    TaskTypes.WAIT,
+                    "externaldebsign",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WORKER,
+                    "makesourcepackageupload",
+                ),
+                (
+                    TaskTypes.SERVER,
+                    "packageupload",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WAIT,
+                    "externaldebsign",
+                ),
+            ]
+            + [
+                (
+                    TaskTypes.WAIT,
+                    "externaldebsign",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WORKER,
+                    "sbuild",
+                ),
+                (
+                    TaskTypes.SERVER,
+                    "packageupload",
+                    WorkRequest.Statuses.BLOCKED,
+                    TaskTypes.WAIT,
+                    "externaldebsign",
+                ),
+            ]
+            * 2,
+        )
+
+    def test_get_label(self) -> None:
         """Test get_label."""
         w = self.create_debian_pipeline_workflow(extra_task_data={})
         self.assertEqual(w.get_label(), "run Debian pipeline")
